# *****************************************************************************
# Copyright (c) 2014 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   David Parker
#   Lokesh Haralakatta
# *****************************************************************************

import json
import re
import pytz
import uuid
import threading
import requests
import paho.mqtt.client as paho

from datetime import datetime

from ibmiotf import AbstractClient, HttpAbstractClient, InvalidEventException, UnsupportedAuthenticationMethod,ConfigurationException, ConnectionException, MissingMessageEncoderException,MissingMessageDecoderException
from ibmiotf.codecs import jsonCodec, jsonIotfCodec, xmlCodec


# Support Python 2.7 and 3.4 versions of configparser
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

COMMAND_RE = re.compile("iot-2/cmd/(.+)/fmt/(.+)")


class Command:
    def __init__(self, pahoMessage, messageEncoderModules):
        result = COMMAND_RE.match(pahoMessage.topic)
        if result:
            self.command = result.group(1)
            self.format = result.group(2)

            if self.format in messageEncoderModules:
                message = messageEncoderModules[self.format].decode(pahoMessage)
                self.timestamp = message.timestamp
                self.data = message.data
            else:
                raise MissingMessageDecoderException(self.format)
        else:
            raise InvalidEventException("Received command on invalid topic: %s" % (pahoMessage.topic))


class Client(AbstractClient):

    COMMAND_TOPIC = "iot-2/cmd/+/fmt/+"

    def __init__(self, options, logHandlers=None):
        self._options = options

        ### DEFAULTS ###
        if "domain" not in self._options:
            # Default to the domain for the public cloud offering
            self._options['domain'] = "internetofthings.ibmcloud.com"
        if "clean-session" not in self._options:
            self._options['clean-session'] = "true"

        if "org" not in self._options:
            # Default to the quickstart
            self._options['org'] = "quickstart"

        if "port" not in self._options and self._options["org"] != "quickstart":
            self._options["port"] = 8883;

        if self._options["org"] == "quickstart":
            self._options["port"] = 1883;

        ### REQUIRED ###
        if self._options['org'] == None:
            raise ConfigurationException("Missing required property: org")
        if self._options['type'] == None:
            raise ConfigurationException("Missing required property: type")
        if self._options['id'] == None:
            raise ConfigurationException("Missing required property: id")

        if self._options['org'] != "quickstart":
            if self._options['auth-method'] == None:
                raise ConfigurationException("Missing required property: auth-method")

            if (self._options['auth-method'] == "token"):
                if self._options['auth-token'] == None:
                    raise ConfigurationException("Missing required property for token based authentication: auth-token")
            else:
                raise UnsupportedAuthenticationMethod(options['auth-method'])

        AbstractClient.__init__(
            self,
            domain = self._options['domain'],
            organization = self._options['org'],
            clientId = "d:" + self._options['org'] + ":" + self._options['type'] + ":" + self._options['id'],
            username = "use-token-auth" if (self._options['auth-method'] == "token") else None,
            password = self._options['auth-token'],
            logHandlers = logHandlers,
            cleanSession = self._options['clean-session'],
            port = self._options['port']
        )

        # Add handler for commands if not connected to QuickStart
        if self._options['org'] != "quickstart":
            self.client.message_callback_add("iot-2/cmd/+/fmt/+", self.__onCommand)

        self.subscriptionsAcknowledged = threading.Event()

        # Initialize user supplied callback
        self.commandCallback = None

        self.client.on_connect = self.on_connect

        self.setMessageEncoderModule('json', jsonCodec)
        self.setMessageEncoderModule('json-iotf', jsonIotfCodec)
        self.setMessageEncoderModule('xml', xmlCodec)


    def on_connect(self, client, userdata, flags, rc):
        '''
        This is called after the client has received a CONNACK message from the broker in response to calling connect().
        The parameter rc is an integer giving the return code:

        0: Success
        1: Refused - unacceptable protocol version
        2: Refused - identifier rejected
        3: Refused - server unavailable
        4: Refused - bad user name or password
        5: Refused - not authorised
        '''
        if rc == 0:
            self.connectEvent.set()
            self.logger.info("Connected successfully: %s" % (self.clientId))
            if self._options['org'] != "quickstart":
                self.__subscribeToCommands()
        elif rc == 5:
            self.logAndRaiseException(ConnectionException("Not authorized: s (%s, %s, %s)" % (self.clientId, self.username, self.password)))
        else:
            self.logAndRaiseException(ConnectionException("Connection failed: RC= %s" % (rc)))

    def publishEvent(self, event, msgFormat, data, qos=0, on_publish=None):
        '''
        Publish an event in IoTF.

        Parameters:
            event - the name of this event
            msgFormat - the format of the data for this event
            data - the data for this event

        Optional paramters:
            qos - the equivalent MQTT semantics of quality of service using the same constants (0, 1 and 2)
            on_publish - a function that will be called when receipt of the publication is confirmed.  This
                         has different implications depending on the qos:
                         qos 0 - the client has asynchronously begun to send the event
                         qos 1 and 2 - the client has confirmation of delivery from IoTF
        '''
        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to send event %s because device is not currently connected")
            return False
        else:
            self.logger.debug("Sending event %s with data %s" % (event, json.dumps(data)))
            topic = 'iot-2/evt/'+event+'/fmt/' + msgFormat

            if msgFormat in self._messageEncoderModules:
                payload = self._messageEncoderModules[msgFormat].encode(data, datetime.now(pytz.timezone('UTC')))

                try:
                    result = self.client.publish(topic, payload=payload, qos=qos, retain=False)
                    if result[0] == paho.MQTT_ERR_SUCCESS:
                        if on_publish is not None:
                            self._messagesLock.acquire()
                            if result[1] in self._onPublishCallbacks:
                                # paho callback beat this thread so call callback inline now
                                del self._onPublishCallbacks[result[1]]
                                on_publish()
                            else:
                                # this thread beat paho callback so set up for call later
                                self._onPublishCallbacks[result[1]] = on_publish
                        return True
                    else:
                        return False
                finally:
                    if on_publish is not None:
                        self._messagesLock.release()
            else:
                raise MissingMessageEncoderException(msgFormat)


    def __subscribeToCommands(self):
        '''
        Subscribe to commands sent to this device.
        '''
        if self._options['org'] == "quickstart":
            self.logger.warning("QuickStart applications do not support commands")
            return False

        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to subscribe to commands because device is not currently connected")
            return False
        else:
            self.client.subscribe(Client.COMMAND_TOPIC, qos=1)
            return True

    def __onCommand(self, client, userdata, pahoMessage):
        '''
        Internal callback for device command messages, parses source device from topic string and
        passes the information on to the registerd device command callback
        '''
        with self._recvLock:
            self.recv = self.recv + 1
        try:
            command = Command(pahoMessage, self._messageEncoderModules)
        except InvalidEventException as e:
            self.logger.critical(str(e))
        else:
            self.logger.debug("Received command '%s'" % (command.command))
            if self.commandCallback: self.commandCallback(command)


class HttpClient(HttpAbstractClient):
    """
    A basic device client with limited capabilies that forgoes an active MQTT connection to the service.
    """

    def __init__(self, options, logHandlers=None):
        self._options = options

        ### DEFAULTS ###
        if "domain" not in self._options:
            # Default to the domain for the public cloud offering
            self._options['domain'] = "internetofthings.ibmcloud.com"
        if "clean-session" not in self._options:
            self._options['clean-session'] = "true"

        ### REQUIRED ###
        if self._options['org'] == None:
            raise ConfigurationException("Missing required property: org")
        if self._options['type'] == None:
            raise ConfigurationException("Missing required property: type")
        if self._options['id'] == None:
            raise ConfigurationException("Missing required property: id")

        if self._options['org'] != "quickstart":
            if self._options['auth-method'] == None:
                raise ConfigurationException("Missing required property: auth-method")

            if (self._options['auth-method'] == "token"):
                if self._options['auth-token'] == None:
                    raise ConfigurationException("Missing required property for token based authentication: auth-token")
            else:
                raise UnsupportedAuthenticationMethod(options['authMethod'])

        HttpAbstractClient.__init__(
            self,
            clientId = "d:" + self._options['org'] + ":" + self._options['type'] + ":" + self._options['id'],
            logHandlers = logHandlers
        )

        self.setMessageEncoderModule('json', jsonCodec)
        self.setMessageEncoderModule('xml', xmlCodec)



    def publishEvent(self, event, msgFormat, data):
        """
        Publish an event over HTTP(s) as given supported format
        Throws a ConnectionException with the message "Server not found" if the client is unable to reach the server
        Otherwise it returns the HTTP status code, (200 - 207 for success)
        """
        self.logger.debug("Sending event %s with data %s" % (event, json.dumps(data)))

        templateUrl = 'https://%s.messaging.%s/api/v0002/device/types/%s/devices/%s/events/%s'

        orgid = self._options['org']
        deviceType = self._options['type']
        deviceId = self._options['id']
        authMethod = "use-token-auth"
        authToken = self._options['auth-token']
        credentials = (authMethod, authToken)

        if orgid == 'quickstart':
            authMethod = None
            authToken = None

        intermediateUrl = templateUrl % (orgid, self._options['domain'], deviceType, deviceId, event)
        self.logger.debug("URL: %s",intermediateUrl)
        try:
            self.logger.debug("Sending event %s with data %s" % (event, json.dumps(data)))

            if msgFormat in self._messageEncoderModules:
                payload = self._messageEncoderModules[msgFormat].encode(data, datetime.now(pytz.timezone('UTC')))
                contentType = self.getContentType(msgFormat)
                response = requests.post(intermediateUrl, auth = credentials, data = payload, headers = {'content-type': contentType})
            else:
                raise MissingMessageEncoderException(msgFormat)

        except Exception as e:
            self.logger.error(e)
            raise e

        if response.status_code >= 300:
            self.logger.warning("Unable to send event: HTTP response code = %s" % (response.status_code))
        return response.status_code



class DeviceInfo(object):
    def __init__(self):
        self.serialNumber = None
        self.manufacturer = None
        self.model = None
        self.deviceClass = None
        self.description = None
        self.fwVersion = None
        self.hwVersion = None
        self.descriptiveLocation = None

    def __str__(self):
        return json.dumps(self.__dict__, sort_keys=True)


class DeviceFirmware(object):
    def __init__(self,version = None,name = None,url = None,verifier = None,state = None,updateStatus = None,updatedDateTime = None):
        self.version = version
        self.name = name
        self.url = url
        self.verifier = verifier
        self.state = state
        self.updateStatus = updateStatus
        self.updatedDateTime = updatedDateTime

    def __str__(self):
        return json.dumps(self.__dict__, sort_keys=True)


class ManagedClient(Client):

    # Publish MQTT topics
    MANAGE_TOPIC = 'iotdevice-1/mgmt/manage'
    UNMANAGE_TOPIC = 'iotdevice-1/mgmt/unmanage'
    UPDATE_LOCATION_TOPIC = 'iotdevice-1/device/update/location'
    ADD_ERROR_CODE_TOPIC = 'iotdevice-1/add/diag/errorCodes'
    CLEAR_ERROR_CODES_TOPIC = 'iotdevice-1/clear/diag/errorCodes'
    NOTIFY_TOPIC = 'iotdevice-1/notify'
    RESPONSE_TOPIC = 'iotdevice-1/response'
    ADD_LOG_TOPIC = 'iotdevice-1/add/diag/log'
    CLEAR_LOG_TOPIC = 'iotdevice-1/clear/diag/log'

    # Subscribe MQTT topics
    DM_RESPONSE_TOPIC = 'iotdm-1/response'
    DM_OBSERVE_TOPIC = 'iotdm-1/observe'
    DM_REBOOT_TOPIC = 'iotdm-1/mgmt/initiate/device/reboot'
    DM_FACTORY_REESET ='iotdm-1/mgmt/initiate/device/factory_reset'
    DM_UPDATE_TOPIC = 'iotdm-1/device/update'
    DM_CANCEL_OBSERVE_TOPIC = 'iotdm-1/cancel'
    DM_FIRMWARE_DOWNLOAD_TOPIC = 'iotdm-1/mgmt/initiate/firmware/download'
    DM_FIRMWARE_UPDATE_TOPIC = 'iotdm-1/mgmt/initiate/firmware/update'
    DME_ACTION_TOPIC = 'iotdm-1/mgmt/custom/#'

    #ResponceCode
    RESPONSECODE_FUNCTION_NOT_SUPPORTED = 501
    RESPONSECODE_ACCEPTED = 202
    RESPONSECODE_INTERNAL_ERROR = 500
    RESPONSECODE_BAD_REQUEST = 400

    UPDATESTATE_IDLE = 0
    UPDATESTATE_DOWNLOADING = 1
    UPDATESTATE_DOWNLOADED = 2
    UPDATESTATE_SUCCESS = 0
    UPDATESTATE_IN_PROGRESS = 1
    UPDATESTATE_OUT_OF_MEMORY = 2
    UPDATESTATE_CONNECTION_LOST = 3
    UPDATESTATE_VERIFICATION_FAILED = 4
    UPDATESTATE_UNSUPPORTED_IMAGE = 5
    UPDATESTATE_INVALID_URI = 6


    def __init__(self, options, logHandlers=None, deviceInfo=None):
        if options['org'] == "quickstart":
            raise Exception("Unable to create ManagedClient instance.  QuickStart devices do not support device management")

        Client.__init__(self, options, logHandlers)
        # TODO: Raise fatal exception if tries to create managed device client for QuickStart

        # Initialize user supplied callback
        self.deviceActionCallback = None
        self.firmwereActionCallback  = None
        self.dmeActionCallback = None

        # Add handler for supported device management commands
        self.client.message_callback_add("iotdm-1/#", self.__onDeviceMgmtResponse)
        self.client.message_callback_add(ManagedClient.DM_REBOOT_TOPIC, self.__onRebootRequest)
        self.client.message_callback_add(ManagedClient.DM_FACTORY_REESET, self.__onFactoryResetRequest)
        self.client.message_callback_add(ManagedClient.DM_FIRMWARE_UPDATE_TOPIC,self.__onFirmwereUpdate)
        self.client.message_callback_add(ManagedClient.DM_OBSERVE_TOPIC,self.__onFirmwereObserve)
        self.client.message_callback_add(ManagedClient.DM_FIRMWARE_DOWNLOAD_TOPIC,self.__onFirmwereDownload)
        self.client.message_callback_add(ManagedClient.DM_UPDATE_TOPIC,self.__onUpdatedDevice)
        self.client.message_callback_add(ManagedClient.DM_CANCEL_OBSERVE_TOPIC,self.__onFirmwereCancel)
        self.client.message_callback_add(ManagedClient.DME_ACTION_TOPIC,self.__onDMEActionRequest)

        self.client.on_subscribe = self.on_subscribe

        self.readyForDeviceMgmt = threading.Event()

        # List of DM requests that have not received a response yet
        self._deviceMgmtRequestsPendingLock = threading.Lock()
        self._deviceMgmtRequestsPending = {}

        # List of DM notify hook
        self._deviceMgmtObservationsLock = threading.Lock()
        self._deviceMgmtObservations = []

        # Initialize local device data model
        self.metadata = {}
        if deviceInfo is not None:
            self._deviceInfo = deviceInfo
        else:
            self._deviceInfo = DeviceInfo()

        self._location = None
        self._errorCode = None
        self.__firmwareUpdate = None

        self.manageTimer = None

    def setSerialNumber(self, serialNumber):
        self._deviceInfo.serialNumber = serialNumber
        return self.notifyFieldChange("deviceInfo.serialNumber", serialNumber)

    def setManufacturer(self, manufacturer):
        self._deviceInfo.serialNumber = manufacturer
        return self.notifyFieldChange("deviceInfo.manufacturer", manufacturer)

    def setModel(self, model):
        self._deviceInfo.serialNumber = model
        return self.notifyFieldChange("deviceInfo.model", model)

    def setdeviceClass(self, deviceClass):
        self._deviceInfo.deviceClass = deviceClass
        return self.notifyFieldChange("deviceInfo.deviceClass", deviceClass)

    def setDescription(self, description):
        self._deviceInfo.description = description
        return self.notifyFieldChange("deviceInfo.description", description)

    def setFwVersion(self, fwVersion):
        self._deviceInfo.fwVersion = fwVersion
        return self.notifyFieldChange("deviceInfo.fwVersion", fwVersion)

    def setHwVersion(self, hwVersion):
        self._deviceInfo.hwVersion = hwVersion
        return self.notifyFieldChange("deviceInfo.hwVersion", hwVersion)

    def setDescriptiveLocation(self, descriptiveLocation):
        self._deviceInfo.descriptiveLocation = descriptiveLocation
        return self.notifyFieldChange("deviceInfo.descriptiveLocation", descriptiveLocation)


    def notifyFieldChange(self, field, value):
        with self._deviceMgmtObservationsLock:
            if field in self._deviceMgmtObservations:
                if not self.readyForDeviceMgmt.wait(timeout=10):
                    self.logger.warning("Unable to notify service of field change because device is not ready for device management")
                    return threading.Event().set()

                reqId = str(uuid.uuid4())
                message = {
                    "d": {
                        "field": field,
                        "value": value
                    },
                    "reqId": reqId
                }

                resolvedEvent = threading.Event()
                self.client.publish(ManagedClient.NOTIFY_TOPIC, payload=json.dumps(message), qos=1, retain=False)
                with self._deviceMgmtRequestsPendingLock:
                    self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.NOTIFY_TOPIC, "message": message, "event": resolvedEvent}

                return resolvedEvent
            else:
                return threading.Event().set()
    '''
    This is called after the client has received a CONNACK message from the broker in response to calling connect().
    The parameter rc is an integer giving the return code:
    0: Success
    1: Refused - unacceptable protocol version
    2: Refused - identifier rejected
    3: Refused - server unavailable
    4: Refused - bad user name or password
    5: Refused - not authorised
    '''
    def on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.connectEvent.set()
            self.logger.info("Connected successfully: %s, Port: %s" % (self.clientId,self.port))
            if self._options['org'] != "quickstart":
                self.client.subscribe( [(ManagedClient.DM_RESPONSE_TOPIC, 1), (ManagedClient.DM_OBSERVE_TOPIC, 1),
                (ManagedClient.DM_REBOOT_TOPIC,1),(ManagedClient.DM_FACTORY_REESET,1),(ManagedClient.DM_UPDATE_TOPIC,1),
                (ManagedClient.DM_FIRMWARE_UPDATE_TOPIC,1),(ManagedClient.DM_FIRMWARE_DOWNLOAD_TOPIC,1),
                (ManagedClient.DM_CANCEL_OBSERVE_TOPIC,1),(Client.COMMAND_TOPIC, 1),(ManagedClient.DME_ACTION_TOPIC,1)] )
        elif rc == 5:
            self.logAndRaiseException(ConnectionException("Not authorized: s (%s, %s, %s)" % (self.clientId, self.username, self.password)))
        else:
            self.logAndRaiseException(ConnectionException("Connection failed: RC= %s" % (rc)))


    def on_subscribe(self, client, userdata, mid, granted_qos):
        # Once IoTF acknowledges the subscriptions we are able to process commands and responses from device management server
        self.subscriptionsAcknowledged.set()
        self.manage()


    def manage(self, lifetime=3600, supportDeviceActions=True, supportFirmwareActions=True,
                    supportDeviceMgmtExtActions=False, bundleIds=[]):
        # TODO: throw an error, minimum lifetime this client will support is 1 hour, but for now set lifetime to infinite if it's invalid
        if lifetime < 3600:
            lifetime = 0

        if not self.subscriptionsAcknowledged.wait(timeout=10):
            self.logger.warning("Unable to send register for device management because device subscriptions are not in place")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
                    'd': {
                            "lifetime": lifetime,
                            "supports": {
                                            "deviceActions": supportDeviceActions,
                                            "firmwareActions": supportFirmwareActions,
                                        },
                            "deviceInfo" : self._deviceInfo.__dict__,
                            "metadata" : self.metadata
                        },
                    'reqId': reqId
                }
        if supportDeviceMgmtExtActions and len(bundleIds) > 0:
            for bundleId in bundleIds:
                message['d']['supports'][bundleId] = supportDeviceMgmtExtActions

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.MANAGE_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.MANAGE_TOPIC, "message": message, "event": resolvedEvent}

        # Register the future call back to Watson IoT Platform 2 minutes before the device lifetime expiry
        if lifetime != 0:
            if self.manageTimer is not None:
                self._logger.debug("Cancelling existing manage timer")
                self.manageTimer.cancel()
            self.manageTimer = threading.Timer(lifetime-120, self.manage,
            [lifetime, supportDeviceActions, supportFirmwareActions, supportDeviceMgmtExtActions, bundleIds]).start()

        return resolvedEvent


    def unmanage(self):
        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to set device to unmanaged because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            'reqId': reqId
        }

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.UNMANAGE_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.UNMANAGE_TOPIC, "message": message, "event": resolvedEvent}

        return resolvedEvent

    def setLocation(self, longitude, latitude, elevation=None, accuracy=None):
        # TODO: Add validation (e.g. ensure numeric values)
        if self._location is None:
            self._location = {}

        self._location['longitude'] = longitude
        self._location['latitude'] = latitude
        if elevation:
            self._location['elevation'] = elevation

        self._location['measuredDateTime'] = datetime.now(pytz.timezone('UTC')).isoformat()

        if accuracy:
            self._location['accuracy'] = accuracy
        elif "accuracy" in self._location:
            del self._location["accuracy"]

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to publish device location because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "d": self._location,
            "reqId": reqId
        }

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.UPDATE_LOCATION_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.UPDATE_LOCATION_TOPIC, "message": message, "event": resolvedEvent}

        return resolvedEvent


    def setErrorCode(self, errorCode=0):
        if errorCode is None:
            errorCode = 0;

        self._errorCode = errorCode

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to publish error code because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "d": { "errorCode": errorCode },
            "reqId": reqId
        }

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.ADD_ERROR_CODE_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.ADD_ERROR_CODE_TOPIC, "message": message, "event": resolvedEvent}

        return resolvedEvent

    def clearErrorCodes(self):
        self._errorCode = None

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to clear error codes because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "reqId": reqId
        }

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.CLEAR_ERROR_CODES_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.CLEAR_ERROR_CODES_TOPIC, "message": message, "event": resolvedEvent}

        return resolvedEvent

    def addLog(self, msg="",data="",sensitivity=0):
        timestamp = datetime.now().isoformat()
        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to publish error code because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "d": {
                "message": msg,
                "timestamp": timestamp,
                "data": data,
                "severity": sensitivity
                 },
            "reqId": reqId
        }

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.ADD_LOG_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.ADD_LOG_TOPIC, "message": message, "event": resolvedEvent}

        return resolvedEvent

    def clearLog(self):

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to clear log because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "reqId": reqId
        }

        resolvedEvent = threading.Event()
        self.client.publish(ManagedClient.CLEAR_LOG_TOPIC, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": ManagedClient.CLEAR_LOG_TOPIC, "message": message, "event": resolvedEvent}

        return resolvedEvent


    def __onDeviceMgmtResponse(self, client, userdata, pahoMessage):

        with self._recvLock:
            self.recv = self.recv + 1

        try:
            data = json.loads(pahoMessage.payload.decode("utf-8"))
            if 'rc' not in data:
                return True
            rc = data['rc']
            reqId = data['reqId']
        except ValueError as e:
            raise Exception("Unable to parse JSON.  payload=\"%s\" error=%s" % (pahoMessage.payload, str(e)))
        else:
            request = None
            with self._deviceMgmtRequestsPendingLock:
                try:
                    request = self._deviceMgmtRequestsPending.pop(reqId)
                except KeyError:
                    self.logger.warning("Received unexpected response from device management: %s" % (reqId))
                else:
                    self.logger.debug("Remaining unprocessed device management requests: %s" % (len(self._deviceMgmtRequestsPending)))


            if request is None:
                return False

            if request['topic'] == ManagedClient.MANAGE_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Manage action completed: %s" % (rc, json.dumps(request['message'])))
                    self.readyForDeviceMgmt.set()
                else:
                    self.logger.critical("[%s] Manage action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == ManagedClient.UNMANAGE_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Unmanage action completed: %s" % (rc, json.dumps(request['message'])))
                    self.readyForDeviceMgmt.clear()
                else:
                    self.logger.critical("[%s] Unmanage action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == ManagedClient.UPDATE_LOCATION_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Location update action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Location update action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == ManagedClient.ADD_ERROR_CODE_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Add error code action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Add error code action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == ManagedClient.CLEAR_ERROR_CODES_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Clear error codes action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Clear error codes action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == ManagedClient.ADD_LOG_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Add log action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Add log action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == ManagedClient.CLEAR_LOG_TOPIC:
                if rc == 200:
                    self.logger.info("[%s] Clear log action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Clear log action failed: %s" % (rc, json.dumps(request['message'])))
            else:
                self.logger.warning("[%s] Unknown action response: %s" % (rc, json.dumps(request['message'])))

            # Now clear the event, allowing anyone that was waiting on this to proceed
            request["event"].set()
            return True

    #Device Action Handlers
    def __onRebootRequest(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_REBOOT_TOPIC,pahoMessage.payload.decode("utf-8")))
        try:
            data = json.loads(pahoMessage.payload.decode("utf-8"))
            reqId = data['reqId']
            if self.deviceActionCallback : self.deviceActionCallback(reqId,"reboot")
        except ValueError as e:
            raise Exception("Unable to process Reboot request.  payload=\"%s\" error=%s" % (pahoMessage.payload, str(e)))

    def __onFactoryResetRequest(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_FACTORY_REESET,pahoMessage.payload.decode("utf-8")))
        try:
            data = json.loads(pahoMessage.payload.decode("utf-8"))
            reqId = data['reqId']
            if self.deviceActionCallback : self.deviceActionCallback(reqId,"reset")
        except ValueError as e:
            raise Exception("Unable to process Factory Reset request.  payload=\"%s\" error=%s" % (pahoMessage.payload, str(e)))

    def respondDeviceAction(self,reqId,responseCode=202,message=""):
        response ={
                        "rc": responseCode,
                        "message": message,
                        "reqId": reqId
                }
        payload=json.dumps(response)
        self.logger.info("Publishing Device Action response with payload :%s" % payload)
        self.client.publish('iotdevice-1/response', payload,qos=1, retain=False)

    #Firmware Handlers
    def __onFirmwereDownload(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_FIRMWARE_DOWNLOAD_TOPIC,pahoMessage.payload.decode("utf-8")))
        data = json.loads(pahoMessage.payload.decode("utf-8"))
        reqId = data['reqId']
        rc = ManagedClient.RESPONSECODE_ACCEPTED
        msg =""
        if self.__firmwareUpdate.state != ManagedClient.UPDATESTATE_IDLE :
            rc = ManagedClient.RESPONSECODE_BAD_REQUEST
            msg = "Cannot download as the device is not in idle state"
        threading.Thread(target= self.respondDeviceAction,args=(reqId,rc,msg)).start()
        if self.firmwereActionCallback :
            self.firmwereActionCallback("download",self.__firmwareUpdate)


    def __onFirmwereCancel(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_CANCEL_OBSERVE_TOPIC,pahoMessage.payload.decode("utf-8")))
        data = json.loads(pahoMessage.payload.decode("utf-8"))
        reqId = data['reqId']
        threading.Thread(target= self.respondDeviceAction,args=(reqId,200,"")).start()

    def __onFirmwereObserve(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_OBSERVE_TOPIC,pahoMessage.payload.decode("utf-8")))
        data = json.loads(pahoMessage.payload.decode("utf-8"))
        reqId = data['reqId']
        #TODO : Proprer validation for fields in payload
        threading.Thread(target= self.respondDeviceAction,args=(reqId,200,"")).start()

    def __onUpdatedDevice(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_UPDATE_TOPIC,pahoMessage.payload.decode("utf-8")))
        data = json.loads(pahoMessage.payload.decode("utf-8"))
        if 'reqId' in data :
                        reqId = data['reqId']
                        d=data['d']
                        value = None
                        for obj in d['fields'] :
                                if 'field' in obj :
                                        if obj['field'] == "mgmt.firmware" :
                                                value = obj["value"]
                        if value != None :
                                self.__firmwareUpdate = DeviceFirmware(value['version'],value['name'],value['uri'],value['verifier'],value['state'],value['updateStatus'],value['updatedDateTime'])
                        threading.Thread(target= self.respondDeviceAction,args=(reqId,204,"")).start()
        else:
                        d=data['d']
                        value = None
                        for obj in d['fields'] :
                                if 'field' in obj :
                                        if obj['field'] == "metadata" :
                                                value = obj["value"]
                        if value != None :
                                self.metadata = value

    def setState(self,status):
        notify = {"d":{"fields":[{"field":"mgmt.firmware","value":{"state":status}}]}}
        if self.__firmwareUpdate != None :
            self.__firmwareUpdate.state = status

        self.logger.info("Publishing state Update with payload :%s" % json.dumps(notify))
        threading.Thread(target= self.client.publish,args=('iotdevice-1/notify',json.dumps(notify),1, False)).start()

    def setUpdateStatus(self,status):
        notify = {"d":{"fields":[{"field":"mgmt.firmware","value":{"state":ManagedClient.UPDATESTATE_IDLE,"updateStatus":status}}]}}
        if self.__firmwareUpdate != None :
            self.__firmwareUpdate.state = ManagedClient.UPDATESTATE_IDLE
            self.__firmwareUpdate.updateStatus = status

        self.logger.info("Publishing  Update Status  with payload :%s" % json.dumps(notify))
        threading.Thread(target= self.client.publish,args=('iotdevice-1/notify',json.dumps(notify),1, False)).start()

    def __onFirmwereUpdate(self,client,userdata,pahoMessage):
        self.logger.info("Message received on topic :%s with payload %s" % (ManagedClient.DM_FIRMWARE_UPDATE_TOPIC,pahoMessage.payload.decode("utf-8")))
        data = json.loads(pahoMessage.payload.decode("utf-8"))
        reqId = data['reqId']
        rc = ManagedClient.RESPONSECODE_ACCEPTED
        msg =""
        if self.__firmwareUpdate.state != ManagedClient.UPDATESTATE_DOWNLOADED :
            rc = ManagedClient.RESPONSECODE_BAD_REQUEST
            msg = "Firmware is still not successfully downloaded."
        threading.Thread(target= self.respondDeviceAction,args=(reqId,rc,msg)).start()
        if self.firmwereActionCallback :
            self.firmwereActionCallback("update",self.__firmwareUpdate)

    def __onDMEActionRequest(self,client,userdata,pahoMessage):
        data = json.loads(pahoMessage.payload.decode("utf-8"))
        self.logger.info("Message received on topic :%s with payload %s"
              % (ManagedClient.DME_ACTION_TOPIC,data))
        reqId = data['reqId']
        if self.dmeActionCallback :
            if self.dmeActionCallback(pahoMessage.topic,data,reqId):
                msg = "DME Action successfully completed from Callback"
                threading.Thread(target= self.respondDeviceAction,args=(reqId,200,msg)).start()
            else:
                msg = "Unexpected device error"
                threading.Thread(target= self.respondDeviceAction,args=(reqId,500,msg)).start()

        else :
            threading.Thread(target= self.respondDeviceAction,args=(reqId,501,"Operation not implemented")).start()


def ParseConfigFile(configFilePath):
    parms = configparser.ConfigParser({
        "domain": "internetofthings.ibmcloud.com",
        "port": "8883",
        "clean-session": "true"
    })
    sectionHeader = "device"

    try:
        with open(configFilePath) as f:
            try:
                parms.read_file(f)
            except AttributeError:
                # Python 2.7 support
                # https://docs.python.org/3/library/configparser.html#configparser.ConfigParser.read_file
                parms.readfp(f)

        domain = parms.get(sectionHeader, "domain")
        organization = parms.get(sectionHeader, "org")
        deviceType = parms.get(sectionHeader, "type")
        deviceId = parms.get(sectionHeader, "id")
        authMethod = parms.get(sectionHeader, "auth-method")
        authToken = parms.get(sectionHeader, "auth-token")
        cleanSession = parms.get(sectionHeader, "clean-session")
        port = parms.get(sectionHeader, "port")

    except IOError as e:
        reason = "Error reading device configuration file '%s' (%s)" % (configFilePath,e[1])
        raise ConfigurationException(reason)

    return {'domain': domain, 'org': organization, 'type': deviceType, 'id': deviceId, 'auth-method': authMethod, 'auth-token': authToken, 'clean-session': cleanSession, 'port': int(port)}



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Lokesh Haralakatta  - Initial Contribution
# *****************************************************************************

import ibmiotf.api
import ibmiotf.application
import ibmiotf.gateway
import ibmiotf.device
from ibmiotf import *
from nose.tools import *
from nose import SkipTest


class TestApi:
    logger=None
    org=None
    deviceType=None
    deviceId=None
    authToken=None
    authKey=None
    invalidAuthKey=None

    @classmethod
    def setup_class(self):
        self.logger = logging.getLogger(self.__module__+".TestApi")
        self.logger.setLevel(logging.INFO)

        appConfFile="application.conf"
        options = ibmiotf.application.ParseConfigFile(appConfFile)

        gwayConfFile="gateway.conf"
        gwayOptions = ibmiotf.gateway.ParseConfigFile(gwayConfFile)

        self.org = options['auth-key'][2:8]
        self.deviceType = options['type']
        self.gatewayType = gwayOptions['type']
        self.deviceId = options['id']
        self.authToken = options['auth-token']
        self.authKey = options['auth-key']
        self.deviceGateway = gwayOptions['id']
        self.newDeviceId = "python-api-test-device"
        self.newDeviceType = "python-api-add-type"
        self.invalidAuthKey = self.authKey[0:9]+"xxxxxxxxxx"

    @classmethod
    def teardown_class(self):
        self.logger=None

    @raises(Exception)
    def testMissingOptions(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.api.ApiClient({},self.logger)
        assert_equal(e.exception.msg, 'Missing required property for API key based authentication: auth-key')

    @raises(Exception)
    def testMissingAuthKey(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.api.ApiClient({"org": self.org, "type": self.deviceType, "id": self.deviceId,
                                        "auth-method": "token" , "auth-token": self.authToken },self.logger)
        assert_equal(e.exception.msg, 'Missing required property for API key based authentication: auth-key')

    @raises(Exception)
    def testAuthKeyNone(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.api.ApiClient({"org": self.org, "type": self.deviceType, "id": self.deviceId,
                                        "auth-method": "token", "auth-token": self.authToken, "auth-key":None},self.logger)
        assert_equal(e.exception.msg, 'Missing required property for API key based authentication: auth-key')

    @raises(Exception)
    def testMissingAuthToken(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.api.ApiClient({"org": self.org, "type": self.deviceType, "id": self.deviceId,
                                   "auth-key": self.authKey },self.logger)
        assert_equal(e.exception.msg, 'Missing required property for API key based authentication: auth-token')

    @raises(Exception)
    def testAuthTokenNone(self):
        with assert_raises(ConfigurationException) as e:
            ibmiotf.api.ApiClient({"org":self.org, "type": self.deviceType, "id": self.deviceId,
                                   "auth-token": None, "auth-key":self.authKey},self.logger)
        assert_equal(e.exception.msg, 'Missing required property for API key based authentication: auth-token')

    def testGetOrganizationDetails(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        orgDetails = apiClient.getOrganizationDetails()
        assert_is_instance(orgDetails,dict)
        assert_equal(orgDetails['name'],"Testing Devices Org")

    @raises(Exception)
    def testGetOrganizationDetailsInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "invalid-token", "auth-key": self.authKey},self.logger)
            apiClient.getOrganizationDetails()
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testGetOrganizationDetailsApiKeyNotExist(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getOrganizationDetails()
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDevicesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDevices(parameters=self.deviceId)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testGetDevicesInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "invalid-token", "auth-key": self.authKey},self.logger)
            apiClient.getDevices()
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testGetDevicesInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDevices()
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')


    def testGetDevices(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        devicesDetails = apiClient.getDevices()
        assert_is_instance(devicesDetails,dict)
        deviceIds = list()
        for item in devicesDetails['results']:
            deviceIds.append(item['deviceId'])

        assert_equal(deviceIds.count(self.deviceId),1)
        assert_equal(deviceIds.count(self.deviceGateway),1)

    @raises(Exception)
    def testRegisterDevicesInvalidRequestException(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            devicesList = ["device-1","device-2","device-3"]
            apiClient.registerDevices(devicesList)
        assert_equal(e.exception.msg, 'Invalid request (No body, invalid JSON, unexpected key, bad value)')

    @SkipTest
    def testRegisterDevicesEmptyDeviceId(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        devicesList = [{'typeId' : "Python-Api-Test"}]
        registerResult = apiClient.registerDevices(devicesList)
        assert_false(registerResult[0]['success'])

    @raises(Exception)
    def testRegisterDevicesMaxLimitException(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            devicesList = list()
            for x in range(30):
                item = {'typeId': 'Python-Api-Test-'+str(x)}
                devicesList.append(item)

            apiClient.registerDevices(devicesList)
        assert_equal(e.exception.msg, 'Maximum number of devices exceeded')

    @raises(Exception)
    def testRegisterDevicesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.registerDevices(None)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testRegisterDeviceDuplicate(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.registerDevice(self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'The device already exists')

    @raises(Exception)
    def testRegisterDeviceInvalidRequest(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.registerDevice("invalid-type","invalid-device")
        assert_equal(e.exception.msg, 'Invalid request (No body, invalid JSON, unexpected key, bad value)')

    @raises(Exception)
    def testRegisterDeviceInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.registerDevice(self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testRegisterDeviceInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.registerDevice(self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    def testRegisterGetUpdateRemoveDevice(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        location = {"longitude" : "12.78", "latitude" : "45.90", "elevation" : "2000", "accuracy" : "0",
                    "measuredDateTime" : "2015-10-28T08:45:11.662Z"}
        deviceInfo = {"serialNumber": "001", "manufacturer": "Blueberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
        metadata = {"customField1": "customValue3", "customField2": "customValue4"}

        #Register new device and validate
        registerResult = apiClient.registerDevice(self.deviceType, self.newDeviceId, self.authToken,deviceInfo, location,metadata)
        assert_equal(registerResult['typeId'],self.deviceType)
        assert_equal(registerResult['deviceId'],self.newDeviceId)

        #Get newly registered device
        getResult = apiClient.getDevice(self.deviceType, self.newDeviceId)
        assert_equal(getResult['typeId'],self.deviceType)
        assert_equal(getResult['deviceId'],self.newDeviceId)

        #Update device info for newly registered device
        updDeviceInfo = {"serialNumber": "001", "manufacturer": "Blackberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "ISL Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
        updResult = apiClient.updateDevice(self.deviceType, self.newDeviceId,metadata,updDeviceInfo)
        assert_equal(updResult['typeId'],self.deviceType)
        assert_equal(updResult['deviceId'],self.newDeviceId)
        assert_equal(updResult['deviceInfo']['manufacturer'],"Blackberry")
        assert_equal(updResult['deviceInfo']['descriptiveLocation'],"ISL Bangalore")

        #Remove the newly registered device
        assert_true(apiClient.removeDevice(self.deviceType, self.newDeviceId))

    @raises(Exception)
    def testGetDeviceInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.getDevice(self.deviceType,self.newDeviceId)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testGetDeviceInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDevice(self.deviceType,self.newDeviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDeviceNotExists(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDevice(self.deviceType,"Device-Not-Exists")
        assert_equal(e.exception.msg, 'The device does not exist')

    @raises(Exception)
    def testUpdateDeviceInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            metadata = {"customField1": "customValue3", "customField2": "customValue4"}
            apiClient.updateDevice(self.deviceType,self.newDeviceId,metadata)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testUpdateDeviceInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            metadata = {"customField1": "customValue3", "customField2": "customValue4"}
            apiClient.updateDevice(self.deviceType,self.newDeviceId,metadata)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testUpdateDeviceNotExists(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            metadata = {"customField1": "customValue3", "customField2": "customValue4"}
            apiClient.updateDevice(self.deviceType,"Device-Not-Exists",metadata)
        assert_equal(e.exception.msg, 'The device does not exist')

    @raises(Exception)
    def testUpdateDeviceUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            deviceInfo = {"serialNumber": "001", "manufacturer": "Blackberry"}
            apiClient.updateDevice(self.deviceType,self.newDeviceId,self.authToken,deviceInfo)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testRemoveDeviceInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.removeDevice(self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testRemoveDeviceInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.removeDevice(self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testRemoveDeviceNotExists(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.removeDevice(self.deviceType,"Device-Not-Exists")
        assert_equal(e.exception.msg, 'The device does not exist')

    @raises(Exception)
    def testGetDeviceTypesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceTypes(self.gatewayType)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testGetDeviceTypesInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.getDeviceTypes()
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testGetDeviceTypesInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDeviceTypes()
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    def testGetDeviceTypes(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        typesDetails = apiClient.getDeviceTypes()
        assert_is_instance(typesDetails,dict)
        typesIds = list()
        for item in typesDetails['results']:
            typesIds.append(item['id'])

        assert_equal(typesIds.count(self.deviceType),1)
        assert_equal(typesIds.count(self.gatewayType),1)

    def testAddGetUpdateDeleteDeviceType(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        deviceInfo = {"serialNumber": "001", "manufacturer": "Blueberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
        metadata = {"customField1": "customValue3", "customField2": "customValue4"}

        #Add new device type
        addResult = apiClient.addDeviceType(self.newDeviceType,"Added by apiClient.addDeviceType Method",deviceInfo,metadata)
        assert_equal(addResult['id'],self.newDeviceType)

        #Get newly added device type
        getResult = apiClient.getDeviceType(self.newDeviceType)
        assert_equal(getResult['id'],self.newDeviceType)

        #Update newly added device type
        upd_deviceInfo = {"serialNumber": "001", "manufacturer": "Blackberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
        updResult=apiClient.updateDeviceType(self.newDeviceType, "Added by apiClient.addDeviceType Method",upd_deviceInfo, metadata)
        assert_equal(updResult['deviceInfo']['manufacturer'],"Blackberry")

        #Remove newly added device type
        assert_true(apiClient.deleteDeviceType(self.newDeviceType))

    @raises(Exception)
    def testAddDeviceTypeDuplicate(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.addDeviceType(self.deviceType)
        assert_equal(e.exception.msg, 'The device type already exists')

    @raises(Exception)
    def testAddDeviceTypeInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.addDeviceType(self.newDeviceType)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testAddDeviceTypeInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.addDeviceType(self.newDeviceType)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testAddDeviceTypeInvalidRequest(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.addDeviceType(self.newDeviceType,{'description':'newDescription'})
        assert_equal(e.exception.msg, 'Invalid request (No body, invalid JSON, unexpected key, bad value)')

    @raises(Exception)
    def testGetDeviceTypeNotExists(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceType(None)
        assert_equal(e.exception.msg, 'The device type does not exist')

    @raises(Exception)
    def testGetDeviceTypeInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.getDeviceType(self.newDeviceType)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testGetDeviceTypeInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDeviceType(self.newDeviceType)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testUpdateDeviceTypeNotExists(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            deviceInfo = {"serialNumber": "001", "manufacturer": "Blueberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
            apiClient.updateDeviceType("type-not-exists","no-decription",deviceInfo)
        assert_equal(e.exception.msg, 'The device type does not exist')

    @raises(Exception)
    def testUpdateDeviceTypeInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            deviceInfo = {"serialNumber": "001", "manufacturer": "Blueberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
            apiClient.updateDeviceType(self.newDeviceType,"no-description",deviceInfo)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testUpdateDeviceTypeInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            deviceInfo = {"serialNumber": "001", "manufacturer": "Blueberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
            apiClient.updateDeviceType(self.newDeviceType,"no-description",deviceInfo)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testDeleteDeviceTypeInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.deleteDeviceType(self.newDeviceType)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testDeleteDeviceTypeInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.deleteDeviceType(self.newDeviceType)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testDeleteDeviceInvalidAuthToken(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": "x9lB6C9QOb9Hkp0xS", "auth-key": self.authKey},self.logger)
            apiClient.deleteDevice(self.deviceType,self.newDeviceId)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testDeleteDeviceInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.deleteDevice(self.deviceType,self.newDeviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testDeleteDeviceUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.deleteDevice(self.deviceId,self.deviceType)
        assert_equal(e.exception.msg, 'Unexpected error')

    def testDeleteMulitpleDevices(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)

        assert_true(apiClient.deleteMultipleDevices([{}])[0]['success'])

    @raises(Exception)
    def testDeleteMultipleDevicesInvalidBody(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.deleteMultipleDevices(self.deviceType)
        assert_equal(e.exception.msg, 'Invalid request (No body, invalid JSON, unexpected key, bad value)')

    @raises(Exception)
    def testDeleteMultipleDevicesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.deleteMultipleDevices(None)
        assert_equal(e.exception.msg, 'Unexpected error')

    @SkipTest
    def testGetLastEvents(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)

        events = apiClient.getLastEvents(self.deviceType,self.deviceId)
        eventId = events[0]['eventId']
        event = apiClient.getLastEvent(self.deviceType,self.deviceId,eventId)
        assert_equal(event['eventId'],eventId)

    @raises(Exception)
    def testGetLastEventUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getLastEvent({"typeId":self.deviceType}, {"deviceId":self.deviceId}, "eventId")
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testGetLastEventsUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getLastEvent({"typeId":self.deviceType}, {"deviceId":self.deviceId})
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testGetLastEventNotFoundError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getLastEvent(self.deviceType, self.deviceId, "eventId")
        assert_equal(e.exception.msg, 'Event not found')

    def testDeviceModelMethods(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        location = {"longitude" : 12.78, "latitude" : 45.90, "elevation" : 2000, "accuracy" : 0,
                    "measuredDateTime" : "2015-10-28T08:45:11.662Z"}
        deviceInfo = {"serialNumber": "001", "manufacturer": "Blueberry", "model": "e2", "deviceClass": "A",
                      "descriptiveLocation" : "Bangalore", "fwVersion" : "1.0.1", "hwVersion" : "12.01"}
        metadata = {"customField1": "customValue3", "customField2": "customValue4"}

        #Register new device and validate
        registerResult = apiClient.registerDevice(self.deviceType, self.newDeviceId, self.authToken,deviceInfo, location,metadata)
        assert_equal(registerResult['typeId'],self.deviceType)
        assert_equal(registerResult['deviceId'],self.newDeviceId)

        #Get Device location and validtae the results returned
        locationResult = apiClient.getDeviceLocation(self.deviceType,self.newDeviceId)
        assert_true(locationResult['longitude'] == location['longitude'])
        assert_true(locationResult['latitude'] == location['latitude'])
        assert_true(locationResult['elevation'] == location['elevation'])
        assert_true(locationResult['accuracy'] == location['accuracy'])

        #Update Device Location Information
        location = {"longitude" : 13, "latitude" : 46, "elevation" : 3000, "accuracy" : 100,
                    "measuredDateTime" : "2015-10-28T08:45:11.662Z"}
        updateResult = apiClient.updateDeviceLocation(self.deviceType,self.newDeviceId,location)
        assert_true(updateResult['longitude'] == location['longitude'])
        assert_true(updateResult['latitude'] == location['latitude'])
        assert_true(updateResult['elevation'] == location['elevation'])
        assert_true(updateResult['accuracy'] == location['accuracy'])

        #Remove added device using deleteDevice Method
        assert_true(apiClient.deleteDevice(self.deviceType, self.newDeviceId))

    @raises(Exception)
    def testGetDeviceLocationNotFoundError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceLocation(self.deviceId,self.deviceType)
        assert_equal(e.exception.msg, 'Device location information not found')

    @raises(Exception)
    def testGetDeviceLocationUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceLocation({"typeId":self.deviceId},self.deviceType)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testUpdateDeviceLocationNotFoundError(self):
        location = {"longitude" : "13", "latitude" : "46", "elevation" : "3000", "accuracy" : "100",
                    "measuredDateTime" : "2015-10-28T08:45:11.662Z"}
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.updateDeviceLocation(self.deviceType,"device-not-exists",location)
        assert_equal(e.exception.msg, 'Device location information not found')

    @raises(Exception)
    def testUpdateDeviceLocationUnexpectedError(self):
        location = {"longitude" : "13", "latitude" : "46", "elevation" : "3000", "accuracy" : "100",
                    "measuredDateTime" : "2015-10-28T08:45:11.662Z"}
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.updateDeviceLocation({"typeId":self.deviceType},self.deviceId,location)
        assert_equal(e.exception.msg, 'Unexpected error')

    @SkipTest
    def testGetDeviceManagementInformation(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        apiClient.getDeviceManagementInformation(self.deviceType, self.deviceId)

    @raises(Exception)
    def testGetDeviceManagementInformationInvalidAuth(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementInformation(self.deviceType, self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDeviceManagementInformationUnexepctedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementInformation({"typeId":self.deviceType}, self.deviceId)
        assert_equal(e.exception.msg, 'Unexpected error')

    def testGetConnectionLogs(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        apiClient.getConnectionLogs({"typeId":self.deviceType, "deviceId":self.deviceId})

    @raises(Exception)
    def testGetConnectionLogsUnexepctedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getConnectionLogs(self.deviceType)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testGetConnectionLogsInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getConnectionLogs({"typeId":self.deviceType, "deviceId":self.deviceId})
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    def testDeviceDiagnosticsLogs(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        logData = { "message": "sample log entry for python-ut-device","severity": 0,"data": "Diagnostic log entry","timestamp": "2016-05-16T08:44:23.909Z"}

        #Clear all diagnostic logs for the device
        assert_true(apiClient.clearAllDiagnosticLogs(self.deviceType, self.deviceId))

        #Add one diagnostic log
        assert_true(apiClient.addDiagnosticLog(self.deviceType, self.deviceId, logData))

        #Get all diagnostic logs
        logsData = apiClient.getAllDiagnosticLogs(self.deviceType, self.deviceId)
        assert_equal(logsData[0]['data'],"Diagnostic log entry")

    @raises(Exception)
    def testGetAllDiagnosticLogsDeviceNotFoundError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getAllDiagnosticLogs(self.deviceType,"device-not-exists")
        assert_equal(e.exception.msg, 'Device not found')

    @raises(Exception)
    def testGetAllDiagnosticLogs500UnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getAllDiagnosticLogs("device-not-exists",self.deviceType)
        assert_equal(e.exception.msg, '[500] Unexpected error')

    @raises(Exception)
    def testGetAllDiagnosticLogsUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getAllDiagnosticLogs({"typeId":self.deviceType},{"deviceId":self.deviceId})
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testClearAllDiagnosticLogsDeviceNotFoundError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.clearAllDiagnosticLogs(self.deviceType,"device-not-exists")
        assert_equal(e.exception.msg, 'Device not found')

    @raises(Exception)
    def testClearAllDiagnosticLogs500UnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.clearAllDiagnosticLogs("device-not-exists",self.deviceType)
        assert_equal(e.exception.msg, '[500] Unexpected error')

    @raises(Exception)
    def testClearAllDiagnosticLogsUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.clearAllDiagnosticLogs({"typeId":self.deviceType},{"deviceId":self.deviceId})
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testClearAllDiagnosticLogsInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.clearAllDiagnosticLogs(self.deviceType, self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testAddDiagnosticLogDeviceNotFoundError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            logData = { "message": "sample log entry for python-ut-device","severity": 0,"data": "Diagnostic log entry","timestamp": "2016-05-16T08:44:23.909Z"}
            apiClient.addDiagnosticLog(self.deviceType,"device-not-exists",logData)
        assert_equal(e.exception.msg, 'Device not found')

    @raises(Exception)
    def testAddDiagnosticLog500UnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            logData = { "message": "sample log entry for python-ut-device","severity": 0,"data": "Diagnostic log entry","timestamp": "2016-05-16T08:44:23.909Z"}
            apiClient.addDiagnosticLog("device-not-exists",self.deviceType,logData)
        assert_equal(e.exception.msg, '[500] Unexpected error')

    @raises(Exception)
    def testAddDiagnosticLogUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            logData = { "message": "sample log entry for python-ut-device","severity": 0,"data": "Diagnostic log entry","timestamp": "2016-05-16T08:44:23.909Z"}
            apiClient.addDiagnosticLog({"typeId":self.deviceType},{"deviceId":self.deviceId},logData)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testAddDiagnosticLogInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            logData = { "message": "sample log entry for python-ut-device","severity": 0,"data": "Diagnostic log entry","timestamp": "2016-05-16T08:44:23.909Z"}
            apiClient.addDiagnosticLog(self.deviceType, self.deviceId,logData)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    def testDeviceDiagnosticsErrorCodes(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        errorData = { "errorCode": 0,"timestamp": "2016-05-17T07:16:42.312Z"}

        #Clear all error codes for the device
        assert_true(apiClient.clearAllErrorCodes(self.deviceType, self.deviceId))

        #Add one error code
        assert_true(apiClient.addErrorCode(self.deviceType, self.deviceId, errorData))

        #Get all error codes
        assert_equal(apiClient.getAllDiagnosticErrorCodes(self.deviceType, self.deviceId)[0]['errorCode'],errorData['errorCode'])

    @raises(Exception)
    def testAddErrorCodeInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            errorData = { "errorCode": 0,"timestamp": "2016-05-17T07:16:42.312Z"}
            apiClient.addErrorCode(self.deviceType, self.deviceId, errorData)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testAddErrorCode500UnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.addErrorCode(self.deviceType,self.deviceId,None)
        assert_equal(e.exception.msg, '[500] Unexpected error')

    @raises(Exception)
    def testAddErrorCodeUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            errorData = { "errorCode": "","timestamp": "2016-05-17T07:16:42.312Z"}
            apiClient.addErrorCode(self.deviceType,self.deviceId,errorData)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testAddErrorCodeNotFound(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            errorData = { "errorCode": 0,"timestamp": "2016-05-17T07:16:42.312Z"}
            apiClient.addErrorCode(self.deviceType,None,errorData)
        assert_equal(e.exception.msg, 'Error Code not found')

    @raises(Exception)
    def testGetAllDiagnosticErrorCodesInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getAllDiagnosticErrorCodes(self.deviceType, self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetAllDiagnosticErrorCodes500UnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getAllDiagnosticErrorCodes(None,None)
        assert_equal(e.exception.msg, '[500] Unexpected error')

    @raises(Exception)
    def testGetAllDiagnosticErrorCodesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getAllDiagnosticErrorCodes({"typeId":self.deviceType},self.deviceId)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testGetAllDiagnosticErrorCodesNotFound(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getAllDiagnosticErrorCodes(self.deviceType,None)
        assert_equal(e.exception.msg, 'Error Code not found')

    @raises(Exception)
    def testClearAllErrorCodesInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.clearAllErrorCodes(self.deviceType, self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testClearAllErrorCodesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.clearAllErrorCodes({"typeId":self.deviceType},self.deviceId)
        assert_equal(e.exception.msg, 'Unexpected error')

    @raises(Exception)
    def testClearAllErrorCodesNotFound(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.clearAllErrorCodes(self.deviceType,None)
        assert_equal(e.exception.msg, 'Error Code not found')

    def testGetServiceStatus(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)

        statusData = apiClient.getServiceStatus()

    @raises(Exception)
    def testGetServiceStatusUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getServiceStatus()
        assert_equal(e.exception.msg, 'Unexpected Error')

    @raises(Exception)
    def testGetActiveDevicesBadRequest(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getActiveDevices({"dummy":"value"})
        assert_equal(e.exception.msg, 'Bad Request')

    @raises(Exception)
    def testGetActiveDevicesUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getActiveDevices({'start':'2015','end':'2016'})
        assert_equal(e.exception.msg, 'Unexpected Error')

    @SkipTest
    def testGetActiveDevices(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        apiClient.getActiveDevices({'start':'2015-01-01','end':'2016-01-01'})

    def testGetDataTraffic(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        apiClient.getDataTraffic({'start':'2016-01-01','end':'2016-05-01'})

    @raises(Exception)
    def testGetDataTrafficBadRequest(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDataTraffic({'start':'2017-01-01','end':'2016-05-01'})
        assert_equal(e.exception.msg, 'Bad Request')

    def testDeviceManagementAPIs(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        mgmtRequest = {"action": "device/reboot", "parameters": [{"name": "action","value": "reboot" }],
                       "devices": [{ "typeId": self.deviceType, "deviceId": self.deviceId }]}

        #Instantiate required device managed client
        deviceConfFile="device.conf"
        deviceOptions = ibmiotf.device.ParseConfigFile(deviceConfFile)
        deviceInfoObj = ibmiotf.device.DeviceInfo()
        managedClient = ibmiotf.device.ManagedClient(deviceOptions,deviceInfo=deviceInfoObj)
        managedClient.connect()

        #Initialize device management request
        initResult = apiClient.initiateDeviceManagementRequest(mgmtRequest)
        reqId = initResult['reqId']

        #Get device management request using request id
        getResult = apiClient.getDeviceManagementRequest(reqId)
        assert_true(getResult['action'],mgmtRequest['action'])

        #Get device management request status using request id
        getStatusResult = apiClient.getDeviceManagementRequestStatus(reqId)
        results = getStatusResult['results']
        assert_true(results[0]['typeId'],self.deviceType)
        assert_true(results[0]['deviceId'],self.deviceId)

        #Get device management request status using typeId and deviceId
        getStatusResult = apiClient.getDeviceManagementRequestStatusByDevice(reqId,self.deviceType,self.deviceId)
        assert_true(getStatusResult['status'] == 1)
        assert_false(getStatusResult['complete'])

        #Get all management requests
        getResult = apiClient.getAllDeviceManagementRequests()
        results = getResult['results']
        assert_true(results[0]['action'],mgmtRequest['action'])

        #Delete initiated device management request
        assert_true(apiClient.deleteDeviceManagementRequest(reqId))

        #Disconnect managed client
        managedClient.disconnect()

    @raises(Exception)
    def testInitiateDeviceManagementRequestUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.initiateDeviceManagementRequest({'typeId':'Python'})
        assert_equal(e.exception.msg, 'Unexpected Error')

    @raises(Exception)
    def testInitiateDeviceManagementRequest500UnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.initiateDeviceManagementRequest(None)
        assert_equal(e.exception.msg, '500 Unexpected Error')

    @raises(Exception)
    def testDeleteDeviceManagementRequestUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.deleteDeviceManagementRequest({'typeId':'Python'})
        assert_equal(e.exception.msg, 'Unexpected Error')

    @raises(Exception)
    def testDeleteDeviceManagementRequestBadRequestID(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.deleteDeviceManagementRequest('not-exists')
        assert_equal(e.exception.msg, 'Request Id not found')

    @raises(Exception)
    def testDeleteDeviceManagementRequestInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.deleteDeviceManagementRequest('requestId')
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDeviceManagementRequestUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementRequest({'typeId':'Python'})
        assert_equal(e.exception.msg, 'Unexpected Error')

    @raises(Exception)
    def testGetDeviceManagementRequestBadRequestID(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementRequest('not-exists')
        assert_equal(e.exception.msg, 'Request Id not found')

    @raises(Exception)
    def testGetDeviceManagementRequestInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDeviceManagementRequest('requestId')
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDeviceManagementRequestStatusUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementRequestStatus({'typeId':'Python'})
        assert_equal(e.exception.msg, 'Unexpected Error')

    @raises(Exception)
    def testGetDeviceManagementRequestStatusBadRequestID(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementRequestStatus('not-exists')
        assert_equal(e.exception.msg, 'Request Id not found')

    @raises(Exception)
    def testGetDeviceManagementRequestStatusInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDeviceManagementRequestStatus('requestId')
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDeviceManagementRequestStatusByDeviceInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDeviceManagementRequestStatusByDevice('requestId',self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testGetDeviceManagementRequestStatusByDeviceBadRequestID(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementRequestStatusByDevice('id-not-exists',self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'Request status not found')

    @raises(Exception)
    def testGetDeviceManagementRequestStatusByDeviceUnexpectedError(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDeviceManagementRequestStatusByDevice({'reqId':'req-id'},self.deviceType,self.deviceId)
        assert_equal(e.exception.msg, 'Unexpected Error')

    def testGatewayAPISupport(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
        gwTypeId = "add-gw-device-type-test"
        devId = "reg-dev-undr-gw-test"

        #Add new gateway device type
        addResult = apiClient.addGatewayDeviceType(gwTypeId)
        assert_equal(addResult['id'],gwTypeId)

        #Get devices under above added gateway device type
        getResult = apiClient.getDevicesConnectedThroughGateway(gwTypeId)
        assert_equal(getResult['results'],[])

        #Add new device through above added gateway
        addResult = apiClient.registerDeviceUnderGateway(gwTypeId,devId)
        assert_equal(addResult['typeId'],gwTypeId)
        assert_equal(addResult['deviceId'],devId)

        #Get devices under above added gateway device type
        getResult = apiClient.getDevicesConnectedThroughGateway(gwTypeId)
        assert_equal(getResult['results'][0]['typeId'],gwTypeId)
        assert_equal(getResult['results'][0]['deviceId'],devId)

        getResult = apiClient.getDevicesConnectedThroughGateway(gwTypeId,devId)
        assert_equal(getResult['typeId'],gwTypeId)
        assert_equal(getResult['deviceId'],devId)

        #Remove the added device followed by the removal of gateway device type
        assert_true(apiClient.deleteDevice(gwTypeId, devId))
        assert_true(apiClient.deleteDeviceType(gwTypeId))

    @raises(Exception)
    def testgetDevicesConnectedThroughGatewayInvalidAuthKey(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDevicesConnectedThroughGateway(self.gatewayType)
        assert_equal(e.exception.msg, 'The authentication method is invalid or the api key used does not exist')

    @raises(Exception)
    def testgetDevicesConnectedThroughGatewayAuthTokenEmpty(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken[0:-1], "auth-key": self.authKey},self.logger)
            apiClient.getDevicesConnectedThroughGateway(self.gatewayType)
        assert_equal(e.exception.msg, 'The authentication token is empty or invalid')

    @raises(Exception)
    def testgetDevicesConnectedThroughGatewayNotExistingDevice(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken, "auth-key": self.authKey},self.logger)
            apiClient.getDevicesConnectedThroughGateway(self.gatewayType,"device-not-exists")
        assert_equal(e.exception.msg, 'The device does not exist')

    def testDeviceManagementExtensionMethods(self):
        apiClient = ibmiotf.api.ApiClient({"auth-method": "token",
           "auth-token": self.authToken, "auth-key": self.authKey})

        dmeData = {"bundleId": "example-dme-actions-v1",
                   "displayName": {"en_US": "example-dme Actions v1"},
                   "version": "1.0","actions": {"installPlugin": {
                   "actionDisplayName": { "en_US": "Install Plug-in"},
                   "parameters": [ { "name": "pluginURI",
                                     "value": "http://example.dme.com",
                                    "required": "true" } ] } } }
        addResult = apiClient.createDeviceManagementExtensionPkg(dmeData)
        assert_equal(addResult['bundleId'],'example-dme-actions-v1')

        getResult = apiClient.getAllDeviceManagementExtensionPkgs();
        assert_equal(getResult['results'][0]['bundleId'],'edgeanalytics')

        getResult = apiClient.getDeviceManagementExtensionPkg('example-dme-actions-v1')
        assert_equal(getResult['bundleId'],'example-dme-actions-v1')

        updData = {"displayName": {"en_US": "example-dme Actions v1"},
                   "version": "1.1","actions": {"installPlugin": {
                   "actionDisplayName": { "en_US": "Install Plug-in"},
                   "parameters": [ { "name": "pluginURI",
                                     "value": "http://example.dme.com",
                                    "required": "true" } ] } } }
        updResult = apiClient.updateDeviceManagementExtensionPkg('example-dme-actions-v1',updData)
        assert_equal(updResult['bundleId'],'example-dme-actions-v1')
        assert_equal(updResult['version'],'1.1')

        assert_true(apiClient.deleteDeviceManagementExtensionPkg('example-dme-actions-v1'))

    @raises(Exception)
    def testgetAllDeviceManagementExtensionPkgsException(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken,
                                               "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getAllDeviceManagementExtensionPkgs()
        assert_equal(e.exception.msg, 'HTTP Error in getAllDeviceManagementExtensionPkgs')

    @raises(Exception)
    def testcreateDeviceManagementExtensionPkg(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken,
                                               "auth-key": self.invalidAuthKey},self.logger)
            apiClient.createDeviceManagementExtensionPkg(None)
        assert_equal(e.exception.msg, 'HTTP Error in createDeviceManagementExtensionPkg')

    @raises(Exception)
    def testdeleteDeviceManagementExtensionPkg(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken,
                                               "auth-key": self.invalidAuthKey},self.logger)
            apiClient.deleteDeviceManagementExtensionPkg(None)
        assert_equal(e.exception.msg, 'HTTP Error in deleteDeviceManagementExtensionPkg')

    @raises(Exception)
    def testgetDeviceManagementExtensionPkg(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken,
                                               "auth-key": self.invalidAuthKey},self.logger)
            apiClient.getDeviceManagementExtensionPkg(None)
        assert_equal(e.exception.msg, 'HTTP Error in getDeviceManagementExtensionPkg')

    @raises(Exception)
    def testupdateDeviceManagementExtensionPkg(self):
        with assert_raises(APIException) as e:
            apiClient = ibmiotf.api.ApiClient({"auth-method": "token","auth-token": self.authToken,
                                               "auth-key": self.invalidAuthKey})
            apiClient.updateDeviceManagementExtensionPkg(None,None)
        assert_equal(e.exception.msg, 'HTTP Error in updateDeviceManagementExtensionPkg')



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Amit M Mangalvedkar
#   Lokesh K Haralakatta
# *****************************************************************************

import json
import re
import pytz
import uuid
import threading
import requests
import logging
import paho.mqtt.client as paho

from datetime import datetime

from ibmiotf import AbstractClient, InvalidEventException, UnsupportedAuthenticationMethod,ConfigurationException, ConnectionException, MissingMessageEncoderException,MissingMessageDecoderException
from ibmiotf.codecs import jsonCodec, jsonIotfCodec, xmlCodec
from ibmiotf import api

# Support Python 2.7 and 3.4 versions of configparser
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

COMMAND_RE = re.compile("iot-2/type/(.+)/id/(.+)/cmd/(.+)/fmt/(.+)")


class Command:
    def __init__(self, pahoMessage, messageEncoderModules):
        result = COMMAND_RE.match(pahoMessage.topic)
        if result:
            self.type = result.group(1)
            self.id = result.group(2)
            self.command = result.group(3)
            self.format = result.group(4)

            if self.format in messageEncoderModules:
                message = messageEncoderModules[self.format].decode(pahoMessage)
                self.timestamp = message.timestamp
                self.data = message.data
            else:
                raise MissingMessageDecoderException(self.format)
        else:
            raise InvalidEventException("Received command on invalid topic: %s" % (pahoMessage.topic))


class Client(AbstractClient):

    def __init__(self, options, logHandlers=None):
        self._options = options

        #Defaults
        if "domain" not in self._options:
            # Default to the domain for the public cloud offering
            self._options['domain'] = "internetofthings.ibmcloud.com"

        if "org" not in self._options:
            # Default to the quickstart ode
            self._options['org'] = "quickstart"

        if "clean-session" not in self._options:
            self._options['clean-session'] = "true"

        if "port" not in self._options and self._options["org"] != "quickstart":
            self._options["port"] = 8883;

        if self._options["org"] == "quickstart":
            self._options["port"] = 1883;

        #Check for any missing required properties
        if self._options['org'] == None:
            raise ConfigurationException("Missing required property: org")
        if self._options['type'] == None:
            raise ConfigurationException("Missing required property: type")
        if self._options['id'] == None:
            raise ConfigurationException("Missing required property: id")

        if self._options['org'] != "quickstart":
            if self._options['auth-method'] == None:
                raise ConfigurationException("Missing required property: auth-method")

            if (self._options['auth-method'] == "token"):
                if self._options['auth-token'] == None:
                    raise ConfigurationException("Missing required property for token based authentication: auth-token")
            else:
                raise UnsupportedAuthenticationMethod(options['authMethod'])
        self._options['subscriptionList'] = {}

        
        self.COMMAND_TOPIC = "iot-2/type/" + self._options['type'] + "/id/" + self._options['id'] + "/cmd/+/fmt/+"
        
        AbstractClient.__init__(
            self,
            domain = self._options['domain'],
            organization = self._options['org'],
            clientId = "g:" + self._options['org'] + ":" + self._options['type'] + ":" + self._options['id'],
            username = "use-token-auth" if (self._options['auth-method'] == "token") else None,
            password = self._options['auth-token'],
            logHandlers = logHandlers,
            port = self._options['port']
        )


        # Add handler for commands if not connected to QuickStart
        if self._options['org'] != "quickstart":
            gatewayCommandTopic = "iot-2/type/" + options['type'] + "/id/" + options['id'] + "/cmd/+/fmt/json"
            messageNotificationTopic = "iot-2/type/" + options['type'] + "/id/" + options['id'] + "/notify"
            #localTopic = "iot-2/type/iotsample-raspberrypi2/id/89898889/cmd/greeting/fmt/json"
            self.client.message_callback_add(gatewayCommandTopic, self.__onCommand)
            self.client.message_callback_add("iot-2/type/+/id/+/cmd/+/fmt/+", self.__onDeviceCommand)
            self.client.message_callback_add(messageNotificationTopic, self.__onMessageNotification)


        self.subscriptionsAcknowledged = threading.Event()

        # Initialize user supplied callback
        self.commandCallback = None
        self.deviceCommandCallback = None
        self.notificationCallback = None
        self.client.on_connect = self.on_connect
        self.setMessageEncoderModule('json', jsonCodec)
        self.setMessageEncoderModule('json-iotf', jsonIotfCodec)
        self.setMessageEncoderModule('xml', xmlCodec)

        # Create api key for gateway authentication
        self.gatewayApiKey = "g/" + self._options['org'] + '/' + self._options['type'] + '/' + self._options['id']
        self.logger = logging.getLogger(self.__module__+"."+self.__class__.__name__)
        self.logger.setLevel(logging.INFO)
        self.api = api.ApiClient({"org": self._options['org'], "auth-token": self._options['auth-token'], "auth-key": self.gatewayApiKey }, self.logger)

    '''
    This is called after the client has received a CONNACK message from the broker in response to calling connect().
    The parameter rc is an integer giving the return code:
    0: Success
    1: Refused - unacceptable protocol version
    2: Refused - identifier rejected
    3: Refused - server unavailable
    4: Refused - bad user name or password
    5: Refused - not authorised
    '''
    def on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.connectEvent.set()
            self.logger.info("Connected successfully: %s" % (self.clientId))
            #if self._options['org'] != "quickstart":
                #self.subscribeToGatewayCommands()
        elif rc == 5:
            self.logAndRaiseException(ConnectionException("Not authorized: s (%s, %s, %s)" % (self.clientId, self.username, self.password)))
        else:
            self.logAndRaiseException(ConnectionException("Connection failed: RC= %s" % (rc)))


    '''
    Publish an event in Watson IoT.
    Parameters:
        event - the name of this event
        msgFormat - the format of the data for this event
        data - the data for this event
        deviceType - the device type of the device on the behalf of which the gateway is publishing the event

    Optional paramters:
        qos - the equivalent MQTT semantics of quality of service using the same constants (0, 1 and 2)
        on_publish - a function that will be called when receipt of the publication is confirmed.  This
                     has different implications depending on the qos:
                     qos 0 - the client has asynchronously begun to send the event
                     qos 1 and 2 - the client has confirmation of delivery from Watson IoT
    '''
    def publishDeviceEvent(self, deviceType, deviceId, event, msgFormat, data, qos=0, on_publish=None):
        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to send event %s because gateway as a device is not currently connected")
            return False
        else:
            self.logger.debug("Sending event %s with data %s" % (event, json.dumps(data)))
            topic = 'iot-2/type/' + deviceType + '/id/' + deviceId +'/evt/'+event+'/fmt/' + msgFormat

            if msgFormat in self._messageEncoderModules:
                payload = self._messageEncoderModules[msgFormat].encode(data, datetime.now(pytz.timezone('UTC')))

                try:
                    result = self.client.publish(topic, payload=payload, qos=qos, retain=False)
                    if result[0] == paho.MQTT_ERR_SUCCESS:
                        if on_publish is not None:
                            self._messagesLock.acquire()
                            if result[1] in self._onPublishCallbacks:
                                # paho callback beat this thread so call callback inline now
                                del self._onPublishCallbacks[result[1]]
                                on_publish()
                            else:
                                # this thread beat paho callback so set up for call later
                                self._onPublishCallbacks[result[1]] = on_publish
                        return True
                    else:
                        return False
                finally:
                    if on_publish is not None:
                        self._messagesLock.release()
            else:
                raise MissingMessageEncoderException(msgFormat)


    '''
    Publish an event in Watson IoT as a device.
    Parameters:
        event - the name of this event
        msgFormat - the format of the data for this event
        data - the data for this event
    Optional paramters:
        qos - the equivalent MQTT semantics of quality of service using the same constants (0, 1 and 2)
        on_publish - a function that will be called when receipt of the publication is confirmed.  This
                     has different implications depending on the qos:
                     qos 0 - the client has asynchronously begun to send the event
                     qos 1 and 2 - the client has confirmation of delivery from Watson IoT
    '''
    def publishGatewayEvent(self, event, msgFormat, data, qos=0, on_publish=None):
        gatewayType = self._options['type']
        gatewayId = self._options['id']

        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to send event %s because gateway as a device is not currently connected")
            return False
        else:
            self.logger.debug("Sending event %s with data %s" % (event, json.dumps(data)))
            topic = 'iot-2/type/' + gatewayType + '/id/' + gatewayId +'/evt/'+event+'/fmt/' + msgFormat

            if msgFormat in self._messageEncoderModules:
                payload = self._messageEncoderModules[msgFormat].encode(data, datetime.now(pytz.timezone('UTC')))

                try:
                    result = self.client.publish(topic, payload=payload, qos=qos, retain=False)
                    if result[0] == paho.MQTT_ERR_SUCCESS:
                        if on_publish is not None:
                            self._messagesLock.acquire()
                            if result[1] in self._onPublishCallbacks:
                                # paho callback beat this thread so call callback inline now
                                on_publish()
                                del self._onPublishCallbacks[result[1]]
                            else:
                                # this thread beat paho callback so set up for call later
                                self._onPublishCallbacks[result[1]] = on_publish
                        return True
                    else:
                        return False
                finally:
                    if on_publish is not None:
                        self._messagesLock.release()
            else:
                raise MissingMessageEncoderException(msgFormat)


    def subscribeToDeviceCommands(self, deviceType, deviceId, command='+', format='json', qos=1):
        if self._options['org'] == "quickstart":
            self.logger.warning("QuickStart not supported in Gateways")
            return False

        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to subscribe to device commands because gateway is not currently connected")
            return False
        else:
            topic = 'iot-2/type/' + deviceType + '/id/' + deviceId + '/cmd/' + command + '/fmt/' + format
            self.client.subscribe(topic, qos=qos)
            self._options['subscriptionList'][topic] = qos
            return True



    def subscribeToGatewayCommands(self, command='+', format='json', qos=1):
        deviceType = self._options['type']
        deviceId = self._options['id']
        if self._options['org'] == "quickstart":
            self.logger.warning("QuickStart not supported in Gateways")
            return False
        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to subscribe to gateway commands because gateway is not currently connected")
            return False
        else:
            topic = 'iot-2/type/' + deviceType + '/id/' + deviceId + '/cmd/' + command + '/fmt/' + format
            self.client.subscribe(topic)
            self._options['subscriptionList'][topic] = qos
            return True


    def subscribeToGatewayNotifications(self):
        deviceType = self._options['type']
        deviceId = self._options['id']
        if self._options['org'] == "quickstart":
            self.logger.warning("QuickStart not supported in Gateways")
            return False
        if not self.connectEvent.wait(timeout=10):
            self.logger.warning("Unable to subscribe to notifications because gateway is not currently connected")
            return False
        else:
            topic = 'iot-2/type/' + deviceType + '/id/' + deviceId + '/notify'
            self.client.subscribe(topic)
            #self._options['subscriptionList'][topic] = qos
            return True


    '''
    Internal callback for device command messages, parses source device from topic string and
    passes the information on to the registered device command callback
    '''
    def __onCommand(self, client, userdata, pahoMessage):
        with self._recvLock:
            self.recv = self.recv + 1
        try:
            command = Command(pahoMessage, self._messageEncoderModules)
        except InvalidEventException as e:
            self.logger.critical(str(e))
        else:
            self.logger.debug("Received device command '%s'" % (command.command))
            if self.commandCallback: self.commandCallback(command)

    '''
    Internal callback for gateway command messages, parses source device from topic string and
    passes the information on to the registered device command callback
    '''
    def __onDeviceCommand(self, client, userdata, pahoMessage):
        with self._recvLock:
            self.recv = self.recv + 1
        try:
            command = Command(pahoMessage, self._messageEncoderModules)
        except InvalidEventException as e:
            self.logger.critical(str(e))
        else:
            self.logger.debug("Received gateway command '%s'" % (command.command))
            if self.deviceCommandCallback: self.deviceCommandCallback(command)

    '''
    Internal callback for gateway notification messages, parses source device from topic string and
    passes the information on to the registered device command callback
    '''
    def __onMessageNotification(self, client, userdata, pahoMessage):
        with self._recvLock:
            self.recv = self.recv + 1
        try:
            command = Command(pahoMessage, self._messageEncoderModules)
        except InvalidEventException as e:
            self.logger.critical(str(e))
        else:
            self.logger.debug("Received Notification '%s'" % (command.command))
            if self.notificationCallback: self.notificationCallback(command)



class DeviceInfo(object):
    def __init__(self):
        self.serialNumber = None
        self.manufacturer = None
        self.model = None
        self.deviceClass = None
        self.description = None
        self.fwVersion = None
        self.hwVersion = None
        self.descriptiveLocation = None

    def __str__(self):
        return json.dumps(self.__dict__, sort_keys=True)


class ManagedClient(Client):

    # Publish MQTT topics
    '''
    MANAGE_TOPIC = 'iotdevice-1/mgmt/manage'
    UNMANAGE_TOPIC = 'iotdevice-1/mgmt/unmanage'
    UPDATE_LOCATION_TOPIC = 'iotdevice-1/device/update/location'
    ADD_ERROR_CODE_TOPIC = 'iotdevice-1/add/diag/errorCodes'
    CLEAR_ERROR_CODES_TOPIC = 'iotdevice-1/clear/diag/errorCodes'
    NOTIFY_TOPIC = 'iotdevice-1/notify'

    # Subscribe MQTT topics
    DM_RESPONSE_TOPIC = 'iotdm-1/response'
    DM_OBSERVE_TOPIC = 'iotdm-1/observe'
    '''

    MANAGE_TOPIC_TEMPLATE = 'iotdevice-1/type/%s/id/%s/mgmt/manage'
    UNMANAGE_TOPIC_TEMPLATE = 'iotdevice-1/type/%s/id/%s/mgmt/unmanage'
    UPDATE_LOCATION_TOPIC_TEMPLATE = 'iotdevice-1/type/%s/id/%s/device/update/location'
    ADD_ERROR_CODE_TOPIC_TEMPLATE = 'iotdevice-1/type/%s/id/%s/add/diag/errorCodes'
    CLEAR_ERROR_CODES_TOPIC_TEMPLATE = 'iotdevice-1/type/%s/id/%s/clear/diag/errorCodes'
    NOTIFY_TOPIC_TEMPLATE = 'iotdevice-1/type/%s/id/%s/notify'

    # Subscribe MQTT topics
    DM_RESPONSE_TOPIC_TEMPLATE = 'iotdm-1/type/%s/id/%s/response'
    DM_OBSERVE_TOPIC_TEMPLATE = 'iotdm-1/type/%s/id/%s/observe'

    def __init__(self, options, logHandlers=None, deviceInfo=None):
        if options['org'] == "quickstart":
            raise Exception("Unable to create ManagedClient instance.  QuickStart devices do not support device management")

        Client.__init__(self, options, logHandlers)
        # TODO: Raise fatal exception if tries to create managed device client for QuickStart

        # Add handler for supported device management commands
        self.client.message_callback_add("iotdm-1/#", self.__onDeviceMgmtResponse)
        self.client.on_subscribe = self.on_subscribe

        self.readyForDeviceMgmt = threading.Event()

        # List of DM requests that have not received a response yet
        self._deviceMgmtRequestsPendingLock = threading.Lock()
        self._deviceMgmtRequestsPending = {}

        # List of DM notify hook
        self._deviceMgmtObservationsLock = threading.Lock()
        self._deviceMgmtObservations = []

        # Initialize local device data model
        self.metadata = {}
        if deviceInfo is not None:
            self._deviceInfo = deviceInfo
        else:
            self._deviceInfo = DeviceInfo()

        self._location = None
        self._errorCode = None

        self._gatewayType = self._options['type']
        self._gatewayId = self._options['id']


    def setSerialNumber(self, serialNumber):
        self._deviceInfo.serialNumber = serialNumber
        return self.notifyFieldChange("deviceInfo.serialNumber", serialNumber)

    def setManufacturer(self, manufacturer):
        self._deviceInfo.serialNumber = manufacturer
        return self.notifyFieldChange("deviceInfo.manufacturer", manufacturer)

    def setModel(self, model):
        self._deviceInfo.serialNumber = model
        return self.notifyFieldChange("deviceInfo.model", model)

    def setdeviceClass(self, deviceClass):
        self._deviceInfo.deviceClass = deviceClass
        return self.notifyFieldChange("deviceInfo.deviceClass", deviceClass)

    def setDescription(self, description):
        self._deviceInfo.description = description
        return self.notifyFieldChange("deviceInfo.description", description)

    def setFwVersion(self, fwVersion):
        self._deviceInfo.fwVersion = fwVersion
        return self.notifyFieldChange("deviceInfo.fwVersion", fwVersion)

    def setHwVersion(self, hwVersion):
        self._deviceInfo.hwVersion = hwVersion
        return self.notifyFieldChange("deviceInfo.hwVersion", hwVersion)

    def setDescriptiveLocation(self, descriptiveLocation):
        self._deviceInfo.descriptiveLocation = descriptiveLocation
        return self.notifyFieldChange("deviceInfo.descriptiveLocation", descriptiveLocation)


    def notifyFieldChange(self, field, value):
        with self._deviceMgmtObservationsLock:
            if field in self._deviceMgmtObservations:
                if not self.readyForDeviceMgmt.wait(timeout=10):
                    self.logger.warning("Unable to notify service of field change because gateway is not ready for gateway management")
                    return threading.Event().set()

                reqId = str(uuid.uuid4())
                message = {
                    "d": {
                        "field": field,
                        "value": value
                    },
                    "reqId": reqId
                }

                notify_topic = ManagedClient.NOTIFY_TOPIC_TEMPLATE %  (self._gatewayType,self._gatewayId)
                resolvedEvent = threading.Event()

                self.client.publish(notify_topic, payload=json.dumps(message), qos=1, retain=False)
                with self._deviceMgmtRequestsPendingLock:
                    self._deviceMgmtRequestsPending[reqId] = {"topic": notify_topic, "message": message, "event": resolvedEvent}

                return resolvedEvent
            else:
                return threading.Event().set()

    def on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.connectEvent.set()
            self.logger.info("Connected successfully: %s, Port: %s" % (self.clientId,self.port))
            if self._options['org'] != "quickstart":
                dm_response_topic = ManagedClient.DM_RESPONSE_TOPIC_TEMPLATE %  (self._gatewayType,self._gatewayId)
                dm_observe_topic = ManagedClient.DM_OBSERVE_TOPIC_TEMPLATE %  (self._gatewayType,self._gatewayId)
                self.client.subscribe( [(dm_response_topic, 1), (dm_observe_topic, 1), (self.COMMAND_TOPIC, 1)] )
        elif rc == 5:
            self.logAndRaiseException(ConnectionException("Not authorized: s (%s, %s, %s)" % (self.clientId, self.username, self.password)))
        else:
            self.logAndRaiseException(ConnectionException("Connection failed: RC= %s" % (rc)))


    def on_subscribe(self, client, userdata, mid, granted_qos):
        # Once Watson IoT acknowledges the subscriptions we are able to process commands and responses from device management server
        self.subscriptionsAcknowledged.set()
        self.manage()


    def manage(self, lifetime=3600, supportDeviceActions=False, supportFirmwareActions=False):
        # TODO: throw an error, minimum lifetime this client will support is 1 hour, but for now set lifetime to infinite if it's invalid
        if lifetime < 3600:
            lifetime = 0

        if not self.subscriptionsAcknowledged.wait(timeout=10):
            self.logger.warning("Unable to send register for device management because device subscriptions are not in place")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            'd': {
                "lifetime": lifetime,
                "supports": {
                    "deviceActions": supportDeviceActions,
                    "firmwareActions": supportFirmwareActions
                },
                "deviceInfo" : self._deviceInfo.__dict__,
                "metadata" : self.metadata
            },
            'reqId': reqId
        }

        manage_topic = ManagedClient.MANAGE_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
        resolvedEvent = threading.Event()

        self.client.publish(manage_topic, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": manage_topic, "message": message, "event": resolvedEvent}

        # Register the future call back to Watson IoT Platform 2 minutes before the device lifetime expiry
        if lifetime != 0:
            threading.Timer(lifetime-120, self.manage, [lifetime, supportDeviceActions, supportFirmwareActions]).start()

        return resolvedEvent


    def unmanage(self):
        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to set device to unmanaged because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            'reqId': reqId
        }

        unmanage_topic = ManagedClient.UNMANAGE_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
        resolvedEvent = threading.Event()

        self.client.publish(unmanage_topic, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": unmanage_topic, "message": message, "event": resolvedEvent}

        return resolvedEvent

    def setLocation(self, longitude, latitude, elevation=None, accuracy=None):
        # TODO: Add validation (e.g. ensure numeric values)
        if self._location is None:
            self._location = {}

        self._location['longitude'] = longitude
        self._location['latitude'] = latitude
        if elevation:
            self._location['elevation'] = elevation

        self._location['measuredDateTime'] = datetime.now(pytz.timezone('UTC')).isoformat()

        if accuracy:
            self._location['accuracy'] = accuracy
        elif "accuracy" in self._location:
            del self._location["accuracy"]

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to publish device location because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "d": self._location,
            "reqId": reqId
        }

        update_location_topic = ManagedClient.UPDATE_LOCATION_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
        resolvedEvent = threading.Event()

        self.client.publish(update_location_topic, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": update_location_topic, "message": message, "event": resolvedEvent}

        return resolvedEvent


    def setErrorCode(self, errorCode=0):
        if errorCode is None:
            errorCode = 0;

        self._errorCode = errorCode

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to publish error code because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "d": { "errorCode": errorCode },
            "reqId": reqId
        }

        add_error_code_topic = ManagedClient.ADD_ERROR_CODE_TOPIC_TEMPLATE %  (self._gatewayType,self._gatewayId)
        resolvedEvent = threading.Event()

        self.client.publish(add_error_code_topic, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": add_error_code_topic, "message": message, "event": resolvedEvent}

        return resolvedEvent

    def clearErrorCodes(self):
        self._errorCode = None

        if not self.readyForDeviceMgmt.wait(timeout=10):
            self.logger.warning("Unable to clear error codes because device is not ready for device management")
            return threading.Event().set()

        reqId = str(uuid.uuid4())
        message = {
            "reqId": reqId
        }

        clear_error_codes_topic = ManagedClient.CLEAR_ERROR_CODES_TOPIC_TEMPLATE %  (self._gatewayType,self._gatewayId)
        resolvedEvent = threading.Event()

        self.client.publish(clear_error_codes_topic, payload=json.dumps(message), qos=1, retain=False)
        with self._deviceMgmtRequestsPendingLock:
            self._deviceMgmtRequestsPending[reqId] = {"topic": clear_error_codes_topic, "message": message, "event": resolvedEvent}

        return resolvedEvent


    def __onDeviceMgmtResponse(self, client, userdata, pahoMessage):
        with self._recvLock:
            self.recv = self.recv + 1

        try:
            data = json.loads(pahoMessage.payload.decode("utf-8"))

            rc = data['rc']
            reqId = data['reqId']
        except ValueError as e:
            raise Exception("Unable to parse JSON.  payload=\"%s\" error=%s" % (pahoMessage.payload, str(e)))
        else:
            request = None
            with self._deviceMgmtRequestsPendingLock:
                try:
                    request = self._deviceMgmtRequestsPending.pop(reqId)
                except KeyError:
                    self.logger.warning("Received unexpected response from device management: %s" % (reqId))
                else:
                    self.logger.debug("Remaining unprocessed device management requests: %s" % (len(self._deviceMgmtRequestsPending)))


            if request is None:
                return False

            manage_topic = ManagedClient.MANAGE_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
            unmanage_topic = ManagedClient.UNMANAGE_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
            update_location_topic = ManagedClient.UPDATE_LOCATION_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
            add_error_code_topic = ManagedClient.ADD_ERROR_CODE_TOPIC_TEMPLATE % (self._gatewayType,self._gatewayId)
            clear_error_codes_topic = ManagedClient.CLEAR_ERROR_CODES_TOPIC_TEMPLATE %  (self._gatewayType,self._gatewayId)

            if request['topic'] == manage_topic:
                if rc == 200:
                    self.logger.info("[%s] Manage action completed: %s" % (rc, json.dumps(request['message'])))
                    self.readyForDeviceMgmt.set()
                else:
                    self.logger.critical("[%s] Manage action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == unmanage_topic:
                if rc == 200:
                    self.logger.info("[%s] Unmanage action completed: %s" % (rc, json.dumps(request['message'])))
                    self.readyForDeviceMgmt.clear()
                else:
                    self.logger.critical("[%s] Unmanage action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == update_location_topic:
                if rc == 200:
                    self.logger.info("[%s] Location update action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Location update action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == add_error_code_topic:
                if rc == 200:
                    self.logger.info("[%s] Add error code action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Add error code action failed: %s" % (rc, json.dumps(request['message'])))

            elif request['topic'] == clear_error_codes_topic:
                if rc == 200:
                    self.logger.info("[%s] Clear error codes action completed: %s" % (rc, json.dumps(request['message'])))
                else:
                    self.logger.critical("[%s] Clear error codes action failed: %s" % (rc, json.dumps(request['message'])))

            else:
                self.logger.warning("[%s] Unknown action response: %s" % (rc, json.dumps(request['message'])))

            # Now clear the event, allowing anyone that was waiting on this to proceed
            request["event"].set()
            return True



def ParseConfigFile(configFilePath):
    parms = configparser.ConfigParser({"domain": "internetofthings.ibmcloud.com",
                                       "port": "8883","clean-session": "true"})
    sectionHeader = "device"
    try:
        with open(configFilePath) as f:
            try:
                parms.read_file(f)

                domain = parms.get(sectionHeader, "domain", fallback="internetofthings.ibmcloud.com")
                organization = parms.get(sectionHeader, "org", fallback=None)
                deviceType = parms.get(sectionHeader, "type", fallback=None)
                deviceId = parms.get(sectionHeader, "id", fallback=None)
                authMethod = parms.get(sectionHeader, "auth-method", fallback=None)
                authToken = parms.get(sectionHeader, "auth-token", fallback=None)
                cleanSession = parms.get(sectionHeader, "clean-session")
                port = parms.get(sectionHeader, "port")
            except AttributeError:
                # Python 2.7 support
                # https://docs.python.org/3/library/configparser.html#configparser.ConfigParser.read_file
                parms.readfp(f)

                domain = parms.get(sectionHeader, "domain", "internetofthings.ibmcloud.com")
                organization = parms.get(sectionHeader, "org", None)
                deviceType = parms.get(sectionHeader, "type", None)
                deviceId = parms.get(sectionHeader, "id", None)
                authMethod = parms.get(sectionHeader, "auth-method", None)
                authToken = parms.get(sectionHeader, "auth-token", None)
                cleanSession = parms.get(sectionHeader, "clean-session",None)
                port = parms.get(sectionHeader, "port",None)

    except IOError as e:
        reason = "Error reading gateway configuration file '%s' (%s)" % (configFilePath,e[1])
        raise ConfigurationException(reason)

    return {'domain': domain, 'org': organization, 'type': deviceType, 'id': deviceId,
            'auth-method': authMethod, 'auth-token': authToken,
            'clean-session': cleanSession, 'port': port}



# *****************************************************************************
# Copyright (c) 2016 IBM Corporation and other Contributors.
#
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#
# Contributors:
#   Lokesh Haralakatta  - Initial Contribution
# *****************************************************************************

import ibmiotf.gateway
import ibmiotf.application
import uuid
from ibmiotf import *
from nose.tools import *
from nose import SkipTest
import testUtils

class TestGateway(testUtils.AbstractTest):
    registeredDevice = None
    registeredGateway = None
    
    DEVICE_TYPE = "test_device"
    DEVICE_ID = str(uuid.uuid4())
    
    GATEWAY_TYPE = "test_gateway"
    GATEWAY_ID = str(uuid.uuid4())

    
    @classmethod
    def setup_class(self):
        # Register a Device
        try: 
            deviceType = self.setupAppClient.api.getDeviceType(self.DEVICE_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.DEVICE_TYPE)
            else: 
                raise e
        self.registeredDevice = self.setupAppClient.api.registerDevice(self.DEVICE_TYPE, self.DEVICE_ID)
        
        # Register a Gateway
        try: 
            gatewayType = self.setupAppClient.api.getDeviceType(self.GATEWAY_TYPE)
        except APIException as e:
            if e.httpCode == 404:
                deviceType = self.setupAppClient.api.addDeviceType(self.GATEWAY_TYPE, classId = "Gateway")
            else: 
                raise e
        self.registeredGateway = self.setupAppClient.api.registerDevice(self.GATEWAY_TYPE, self.GATEWAY_ID)
        
        self.options={
            "org": self.ORG_ID,
            "type": self.registeredGateway["typeId"],
            "id": self.registeredGateway["deviceId"],
            "auth-method": "token",
            "auth-token": self.registeredGateway["authToken"]
        }
        

    @classmethod
    def teardown_class(self):
        self.setupAppClient.api.deleteDevice(self.DEVICE_TYPE, self.DEVICE_ID)
        self.setupAppClient.api.deleteDevice(self.GATEWAY_TYPE, self.GATEWAY_ID)


    def testManagedGatewayInstance(self):
        managedGateway = ibmiotf.gateway.ManagedClient(self.options)
        assert_is_instance(managedGateway, ibmiotf.gateway.ManagedClient)

    @raises(Exception)
    def testManagedgatewayQSException(self):
        with assert_raises(Exception)as e:
            options={"org": "quickstart", "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method":"None", "auth-token":"None" }
            ibmiotf.gateway.managedClient(options)
        assert_equals(e.exception, Exception)

    def testManagedGatewayConnectException(self):
        badOptions = {"org": self.ORG_ID, "type": self.registeredGateway["typeId"], "id": self.registeredGateway["deviceId"], "auth-method":"token", "auth-token":"xxxxxxxxxxxxxxxxxx" }
        gatewayInfoObj = ibmiotf.gateway.DeviceInfo()
        managedGateway = ibmiotf.gateway.ManagedClient(badOptions, deviceInfo=gatewayInfoObj)
        with assert_raises(ConnectionException) as e:
            managedGateway.connect()

    @SkipTest
    def testManagedGatewayInstanceWithDeviceInfo(self):
        gatewayInfoObj = ibmiotf.gateway.DeviceInfo()
        managedGateway = ibmiotf.gateway.ManagedClient(self.options, deviceInfo=gatewayInfoObj)

        assert_is_instance(managedGateway, ibmiotf.gateway.ManagedClient)

        #Connect managedGateway
        managedGateway.connect()

        #Define device properties to be notified whenever reset
        managedGateway._deviceMgmtObservations = ["deviceInfo.manufacturer", "deviceInfo.descriptiveLocation",
                                                  "deviceInfo.fwVersion", "deviceInfo.model", "deviceInfo.description",
                                                  "deviceInfo.deviceClass", "deviceInfo.hwVersion", "deviceInfo.serialNumber"]

        #Reset managedgateway properties
        managedGateway.setErrorCode(1)
        managedGateway.setLocation(longitude=100, latitude=78, accuracy=100,elevation=45)
        managedGateway.setSerialNumber('iot-pgateway-12345')
        managedGateway.setManufacturer("IBM India Pvt Ltd")
        managedGateway.setModel("2016")
        managedGateway.setdeviceClass("Smart Gateway")
        managedGateway.setDescription("Sample Smart IoT Gateway")
        managedGateway.setFwVersion("1.0")
        managedGateway.setHwVersion("2.0")
        managedGateway.setDescriptiveLocation("ISL Lab Bangalore")

        managedGateway.clearErrorCodes()

        #Disconnect ManagedGateway
        managedGateway.unmanage()
        managedGateway.disconnect()
